#define _BSD_SOURCE

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "marxup.h"

#define LINE_MAX 32768
#define puts(...) fprintf(output, __VA_ARGS__)

void marxup_inline(FILE *, char *, char *);

const char *url = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~:/?#@!$&'()*+,;=";

enum marxup_states {
  VOID,
  PARAGRAPH,
  QUOTE,
  TABLE
};

void cleanputc(FILE *output, char c) {
  switch(c) {
    case '&': puts("&amp;"); break;
    case '<': puts("&lt;"); break;
    case '>': puts("&gt;"); break;
    default: putc(c, output);
  }
}

enum marxup_tags { B, I, TT, BLINK, SUP, SUB };
char* marxup_tags[][2] = {
  { "<b>", "</b>" },
  { "<i>", "</i>" },
  { "<tt>", "</tt>" },
  { "<blink>", "</blink>" },
  { "<sup>", "</sup>" },
  { "<sub>", "</sub>" }
};

void marxup_cut(char *p, char *e, char **t, char **tt, char **u, int *abs) {
  char *pos, *scheme;
  scheme = strstr(p, "://");
  if(scheme && scheme < e)
    *abs = 1;
  *t = *u = p;
  *tt = e;
  while((pos = strstr(*u, ": ")) && pos < e)
    *tt = pos, *u = pos + 2;
}

int marxup_link(FILE *output, char *original) {
  char *text, *textend, *url, *pos;
  char *end = strchr(original, ']');
  int absolute, lower;
  pos = original + 1;
  if(!end) return 0;
  while(isspace(*pos)) {
    pos++;
  }
  if(pos == end) {
    puts("[");
    return 0;
  }
  pos = original + 1;
  absolute = 0, lower = 0;
  marxup_cut(original + 1, end, &text, &textend, &url, &absolute);
  if(getenv("MARXUP_WIKI") && !absolute)
    lower = 1;

  puts("<a rel='nofollow' href='");

  if(getenv("MARXUP_PREFIX") && !absolute)
    puts("%s", getenv("MARXUP_PREFIX")), puts("/");
  while(url < end)
    if(isspace(*url))
      url++;
    else
      lower ? cleanputc(output, tolower(*url++)) : cleanputc(output, *url++);
  puts("'>");
  marxup_inline(output, text, textend);
  puts("</a>");
  return end - original;
}

int marxup_image(FILE *output, char *original) {
  char *text, *textend, *url;
  int absolute, lower;
  long width = 0;

  char *pos = original + 1;
  char *end = strchr(original, '}');
  if(!end) return 0;
  while(isspace(*pos)) {
    pos++;
  }
  if(pos == end) {
    puts("{");
    return 0;
  }
  pos = original + 1;

  char *metaend = pos + strspn(pos, "@ ");
  while(pos < metaend) {
    switch(*pos) {
      case '@':
        pos++;
        if(isdigit(*pos)) {
          width = strtol(pos, &pos, 10);
          continue;
        }
        break;
      default:
        pos++;
    }
  }
  while(isspace(*pos))
    pos++;

  absolute = 0, lower = 0;
  marxup_cut(pos, end, &text, &textend, &url, &absolute);

  if(getenv("MARXUP_WIKI") && !absolute)
    lower = 1;

  puts("<img alt='");
  if(text != url) {
    pos = text;
    while(pos < textend)
      cleanputc(output, *pos++);
    puts("' title='");
    while(text < textend)
      cleanputc(output, *text++);
  }

  puts("' src='");

  if(getenv("MARXUP_IMAGE_PREFIX") && !absolute)
    puts("%s", getenv("MARXUP_IMAGE_PREFIX")), puts("/");
  while(url < end)
    if(isspace(*url))
      puts("%s", "%20"), url++;
    else
      lower ? cleanputc(output, tolower(*url++)) : cleanputc(output, *url++);
  if(width)
    puts("' width='%li", width);
  puts("'>");
  return end - original;
}

int marxup_list(FILE *output, char *pos, int lastdepth) {
  int depth = 0;
  while(*pos == '-' || *pos == '*')
    pos++, depth++;
  int d = depth;
  while(isspace(*pos))
    pos++;
  while(depth < lastdepth)
    puts("</ul>\n"), depth++;
  while(depth > lastdepth)
    puts("<ul>\n"), depth--;
  puts("<li>");
  marxup_inline(output, pos, NULL);
  puts("\n");
  return d;
}

int marxup_tag(FILE *output, char c, int tag, char *original, int recurse) {
  char *pos = original;
  char *end = strchr(pos + 1, c);
  if(end) {
    pos++;
    puts("%s", marxup_tags[tag][0]);
    if(recurse)
      marxup_inline(output, pos, end);
    else
      while(pos < end) cleanputc(output, *pos++);
    puts("%s", marxup_tags[tag][1]);
    return (end - original);
  }
  putc(*pos, output);
  return 0;
}

void marxup_row(FILE *output, char *pos) {
  char *token, *end;
  char *freeme, *work;
  if(!pos)
    return;
  freeme = work = strdup(pos);
  puts("<tr>");
  char *t;
  while((token = strsep(&work, "|\n"))) {
    while(isspace(*token))
      token++;
    t = "td";
    if(*token == '=') {
      t = "th";
      token++;
    }
    puts("<%s>", t);
    end = token;
    while(*end) end++;
    marxup_inline(output, token, end);
    puts("</%s>", t);
  }
  puts("</tr>");
  free(freeme);
}

void marxup_map(FILE *output, char *pos) {
  int rect[4];
  int n = 0;
  while(isspace(*pos)) pos++;
  while(n < 4 && isdigit(*pos)) {
    rect[n] = strtol(pos, &pos, 10); n++; pos++;
  }
  if(n < 4) return;
  while(isspace(*pos)) pos++;
  char *end = pos + strspn(pos, url);
  puts("<area shape='rect'");
  puts(" coords='%d,%d,%d,%d'", rect[0], rect[1], rect[2], rect[3]);
  if(end > pos) {
    puts(" href='");
    while(pos < end)
      cleanputc(output, *pos++);
    puts("'");
  }
  puts(">\n");
}

void marxup_inline(FILE *output, char *pos, char *eos) {
  int escape = 0;
  while(*pos && pos != eos) {
    if(escape && *pos != '&' && *pos != '<' && *pos != '>') {
      putc(*pos++, output);
      escape = 0;
      continue;
    }
    switch(*pos) {
      case '&': puts("&amp;"); break;
      case '>': puts("&gt;"); break;
      case '<':
        pos[1] == '3' ? pos++, puts("♥") : puts("&lt;"); break;
      case '*':
        pos += marxup_tag(output, '*', B, pos, 1); break;
      case '_':
        pos += marxup_tag(output, '_', I, pos, 1); break;
      case '$':
        pos += marxup_tag(output, '$', BLINK, pos, 1); break;
      case '`':
        pos += marxup_tag(output, '`', TT, pos, 0); break;
      case '[':
        pos += marxup_link(output, pos); break;
      case '{':
        pos += marxup_image(output, pos); break;
      case '\\':
        escape = 1; break;
      default:
        putc(*pos, output);
    }
    pos++;
  }
}

void marxup(FILE *input, FILE *output, void (*meta)(FILE *, char *)) {
  int state = VOID;
  int depth = 0;
  int section = 0;

  char *buffer, *pos, *last;

  buffer = malloc(LINE_MAX);

  if(!buffer)
    return;

  while((pos = fgets(buffer, LINE_MAX, input))) {
    last = pos + strlen(pos) - 1;
    while(*last == '\n' || *last == '\r')
      *last-- = '\0';

    while(isspace(*pos))
      pos++;

    if(depth > 0 && *pos != '*' && *pos != '-')
      while(depth)
        puts("\n</ul>\n"), depth--;
    if(state == QUOTE && *pos != '>')
      puts("\n</blockquote>\n"), state = VOID;
    if(state == TABLE && *pos != '|')
      puts("\n</table>"), state = VOID;

    switch(*pos) {
      case '=':
        depth = 0;
        if(getenv("MARXUP_HEADER"))
          depth = strtol(getenv("MARXUP_HEADER"), NULL, 10) - 1;
        while(*pos == '=')
          pos++, depth++;
        depth = depth > 6 ? 6 : depth;
        while(isspace(*pos))
          pos++;
        puts("\n<h%i>", depth);
        while(*pos)
          cleanputc(output, *pos++);
        puts("</h%i>\n", depth);
        depth = 0;
        break;
      case '#':
        if(meta) meta(output, ++pos);
        break;
      case '*':
      case '-':
        depth = marxup_list(output, pos, depth);
        break;
      case '|':
        if(state != TABLE)
          puts("\n<table>");
        marxup_row(output, ++pos);
        state = TABLE;
        break;
      case '<':
        if(strstr(pos, "<<<") == pos) {
          pos += 3;
          while(*pos && isspace(*pos)) pos++;
          if(section) puts("\n</div>\n");
          puts("\n<div class='");
          while(*pos) cleanputc(output, *pos++);
          puts("'>\n");
          section = 1;
        }
        break;
      case '>':
        if(section && strstr(pos, ">>>") == pos) {
          puts("\n</div>\n"); section = 0; break;
        }
        while(isspace(*pos) || *pos == '>')
          pos++;
        if (state != QUOTE)
          puts("\n<blockquote>\n<p>");
        else if(*pos)
          *pos == '-' ? puts("\n<p class='source'>") : puts("\n<br>");
        else
          puts("\n<p>");
        marxup_inline(output, pos, NULL);
        state = QUOTE;
        break;
      case '{':
        if(strstr(pos, "{{{")) {
          pos += 3;
          puts("\n<pre>");
          while(*pos)
            cleanputc(output, *pos++);
          while((pos = fgets(buffer, LINE_MAX, input))) {
            if(strstr(pos, "}}}") == pos)
              break;
            while(*pos)
              cleanputc(output, *pos++);
          }
          puts("</pre>\n");
          break;
        }
      case '\\':
      default:
        if(!*pos) {
          state = VOID;
          continue;
        }
        state == PARAGRAPH ? puts("<br>\n") : puts("<p>");
        marxup_inline(output, pos, NULL);
        puts("\n");
        state = PARAGRAPH;
    }
  }
  if(depth > 0)
    while(depth)
      puts("</ul>\n"), depth--;
  if(state == QUOTE)
    puts("\n</blockquote>\n");
  if(state == TABLE)
    puts("\n</table>\n");
  if(section)
    puts("\n</div>");
  puts("\n");
  free(buffer);
}

